class ExperimentSettingError(Exception):
    pass


class ExperimentRunSettingError(Exception):
    pass


class InvalidFileExtensionError(Exception):
    pass


class JsonEncodingError(Exception):
    pass


class ExperimentNotInitializedError(Exception):
    pass


class ReproductionError(Exception):
    pass


class InputShapeError(Exception):
    pass


class DeviceSettingError(Exception):
    pass


class ModelSettingError(Exception):
    pass


class InvalidPlatformError(Exception):
    pass


class InvalidQunatumDeviceError(Exception):
    pass


class InvalidModelNameError(Exception):
    pass


class InvalidConfigError(Exception):
    pass
